import keyword
from typing import List, Set, Optional, Union, Literal

from .exceptions import (
    ParameterValueError,
    ParameterDefinitionError,
    ParameterNameError,
)
from .types import (
    PrimitiveAllowedTypes,
    ParamType,
    ParameterChoice,
)

UNIT_MAX_LEN = 10
DISPLAY_NAME_MAX_LEN = 30
DESCRIPTION_MAX_LEN = 100


def validate_variable_name_unique(
    variable_name: str, other_variable_names: Set[str]
) -> None:
    """Validate that the given variable name is unique."""
    if isinstance(variable_name, str) and variable_name in other_variable_names:
        raise ParameterNameError(
            f'"{variable_name}" is already defined as a variable name for another parameter.'
            f" All variable names must be unique."
        )


def ensure_display_name(display_name: str) -> str:
    """Validate display name is within the character limit."""
    if not isinstance(display_name, str):
        raise ParameterNameError(
            f"Display name must be a string and at most {DISPLAY_NAME_MAX_LEN} characters."
        )
    if len(display_name) > DISPLAY_NAME_MAX_LEN:
        raise ParameterNameError(
            f'Display name "{display_name}" greater than {DISPLAY_NAME_MAX_LEN} characters.'
        )
    return display_name


def ensure_variable_name(variable_name: str) -> str:
    """Validate variable name is a valid python variable name."""
    if not isinstance(variable_name, str):
        raise ParameterNameError("Variable name must be a string.")
    if not variable_name.isidentifier():
        raise ParameterNameError(
            "Variable name must only contain alphanumeric characters, underscores, and cannot start with a digit."
        )
    if keyword.iskeyword(variable_name):
        raise ParameterNameError("Variable name cannot be a reserved Python keyword.")
    return variable_name


def ensure_description(description: Optional[str]) -> Optional[str]:
    """Validate description is within the character limit."""
    if description is not None:
        if not isinstance(description, str):
            raise ParameterNameError(
                f"Description must be a string and at most {DESCRIPTION_MAX_LEN} characters."
            )
        if len(description) > DESCRIPTION_MAX_LEN:
            raise ParameterNameError(
                f'Description "{description}" greater than {DESCRIPTION_MAX_LEN} characters.'
            )
    return description


def ensure_unit_string_length(unit: Optional[str]) -> Optional[str]:
    """Validate unit is within the character limit."""
    if unit is not None:
        if not isinstance(unit, str):
            raise ParameterNameError(
                f"Unit must be a string and at most {UNIT_MAX_LEN} characters."
            )
        if len(unit) > UNIT_MAX_LEN:
            raise ParameterNameError(
                f'Unit "{unit}" greater than {UNIT_MAX_LEN} characters.'
            )
    return unit


def ensure_value_type(
    value: Union[float, bool, str], parameter_type: type
) -> PrimitiveAllowedTypes:
    """Ensures that the value type coming in from the client matches the given type.

    This does not guarantee that the value will be the correct type for the given parameter, only that any data coming
    in is in the format that we expect. For now, the only transformation it is doing is converting integers represented
    as floating points to integers, and bools represented as 1.0/0.0 to True/False, and floating points represented as
    ints to floats.

    If something is labelled as a type but does not get converted here, that will be caught when it is attempted to be
    set as the parameter value and will raise the appropriate error there.
    """
    validated_value: PrimitiveAllowedTypes = value
    if isinstance(value, float):
        if parameter_type is bool and (value == 0 or value == 1):
            validated_value = bool(value)
        elif parameter_type is int and value.is_integer():
            validated_value = int(value)
    elif (
        isinstance(value, int)
        and not isinstance(value, bool)
        and parameter_type is float
    ):
        validated_value = float(value)
    return validated_value


def ensure_float_value(value: Union[float, int]) -> float:
    """Ensures that if we are expecting a float and receive an int, that will be converted to a float."""
    if not isinstance(value, bool) and isinstance(value, int):
        return float(value)
    return value


def ensure_optional_float_value(value: Optional[Union[float, int]]) -> Optional[float]:
    """Ensures that if we are expecting an optional float and receive an int, that will be converted to a float."""
    if not isinstance(value, bool) and isinstance(value, int):
        return float(value)
    return value


def ensure_float_choices(
    choices: Optional[List[ParameterChoice]],
) -> Optional[List[ParameterChoice]]:
    """Ensures that if we are expecting float parameter choices and any are int types, those will be converted."""
    if choices is not None:
        return [
            ParameterChoice(
                display_name=choice["display_name"],
                # Type ignore because if for some reason this is a str or bool, that will raise in `validate_options`
                value=ensure_float_value(choice["value"]),  # type: ignore[arg-type]
            )
            for choice in choices
        ]
    return choices


def convert_type_string_for_enum(
    parameter_type: type,
) -> Literal["int", "float", "str"]:
    """Converts a type object into a string for an enumerated parameter."""
    if parameter_type is int:
        return "int"
    elif parameter_type is float:
        return "float"
    elif parameter_type is str:
        return "str"
    else:
        raise ParameterValueError(
            f"Cannot resolve parameter type '{parameter_type.__name__}' for an enumerated parameter."
        )


def convert_type_string_for_num_param(parameter_type: type) -> Literal["int", "float"]:
    """Converts a type object into a string for a number parameter."""
    if parameter_type is int:
        return "int"
    elif parameter_type is float:
        return "float"
    else:
        raise ParameterValueError(
            f"Cannot resolve parameter type '{parameter_type.__name__}' for a number parameter."
        )


def _validate_choices(
    minimum: Optional[ParamType],
    maximum: Optional[ParamType],
    choices: List[ParameterChoice],
    parameter_type: type,
) -> None:
    """Validate that min and max is not defined and all choices are properly formatted."""
    if minimum is not None or maximum is not None:
        raise ParameterDefinitionError(
            "If choices are provided minimum and maximum values cannot be provided."
        )
    for choice in choices:
        try:
            display_name = choice["display_name"]
            value = choice["value"]
        except KeyError:
            raise ParameterDefinitionError(
                "All choices must be a dictionary with keys 'display_name' and 'value'."
            )
        ensure_display_name(display_name)
        if not isinstance(value, parameter_type):
            raise ParameterDefinitionError(
                f"All choices provided must be of type '{parameter_type.__name__}'"
            )


def _validate_min_and_max(
    minimum: Optional[ParamType],
    maximum: Optional[ParamType],
    parameter_type: type,
) -> None:
    """Validate the minium and maximum are both defined, the same type, and a valid range."""
    if minimum is not None and maximum is None:
        raise ParameterDefinitionError(
            "If a minimum value is provided a maximum must also be provided."
        )
    elif maximum is not None and minimum is None:
        raise ParameterDefinitionError(
            "If a maximum value is provided a minimum must also be provided."
        )
    elif maximum is not None and minimum is not None:
        if parameter_type is int or parameter_type is float:
            if not isinstance(minimum, parameter_type):
                raise ParameterDefinitionError(
                    f"Minimum is type '{type(minimum).__name__}',"
                    f" but must be of parameter type '{parameter_type.__name__}'"
                )
            if not isinstance(maximum, parameter_type):
                raise ParameterDefinitionError(
                    f"Maximum is type '{type(maximum).__name__}',"
                    f" but must be of parameter type '{parameter_type.__name__}'"
                )
            # These asserts are for the type checker and should never actually be asserted false
            assert isinstance(minimum, (int, float))
            assert isinstance(maximum, (int, float))
            if maximum < minimum:
                raise ParameterDefinitionError(
                    "Maximum must be greater than the minimum"
                )
        else:
            raise ParameterDefinitionError(
                "Only parameters of type float or int can have a minimum and maximum."
            )


def validate_type(value: ParamType, parameter_type: type) -> None:
    """Validate parameter value is the correct type."""
    if not isinstance(value, parameter_type):
        raise ParameterValueError(
            f"Parameter value {value!r} has type '{type(value).__name__}',"
            f" but must be of type '{parameter_type.__name__}'."
        )


def validate_options(
    default: ParamType,
    minimum: Optional[ParamType],
    maximum: Optional[ParamType],
    choices: Optional[List[ParameterChoice]],
    parameter_type: type,
) -> None:
    """Validate default values and all possible constraints for a valid parameter definition."""
    if not isinstance(default, parameter_type):
        raise ParameterValueError(
            f"Parameter default {default!r} has type '{type(default).__name__}',"
            f" but must be of type '{parameter_type.__name__}'."
        )

    if choices is None and minimum is None and maximum is None:
        raise ParameterDefinitionError(
            "Must provide either choices or a minimum and maximum value"
        )

    if choices is not None:
        _validate_choices(minimum, maximum, choices, parameter_type)
    else:
        _validate_min_and_max(minimum, maximum, parameter_type)



"""An emulation of the opentrons thermocycler module.

The purpose is to provide a fake backend that responds to GCODE commands.
"""

import logging
from typing import Optional
from opentrons.drivers.thermocycler.driver import GCODE
from opentrons.drivers.types import ThermocyclerLidStatus
from opentrons.hardware_control.emulation.parser import Parser, Command
from opentrons.hardware_control.emulation.settings import ThermocyclerSettings

from .abstract_emulator import AbstractEmulator
from .simulations import Temperature, TemperatureWithHold
from . import util

logger = logging.getLogger(__name__)


class ThermocyclerEmulator(AbstractEmulator):
    """Thermocycler emulator"""

    _lid_temperature: Temperature
    _plate_temperature: TemperatureWithHold
    lid_status: ThermocyclerLidStatus
    plate_volume: util.OptionalValue[float]
    plate_ramp_rate: util.OptionalValue[float]

    def __init__(self, parser: Parser, settings: ThermocyclerSettings) -> None:
        self._parser = parser
        self._settings = settings
        self.reset()

    def handle(self, line: str) -> Optional[str]:
        """Handle a line"""
        results = (self._handle(c) for c in self._parser.parse(line))
        joined = " ".join(r for r in results if r)
        return None if not joined else joined

    def reset(self) -> None:
        self._lid_temperature = Temperature(
            per_tick=self._settings.lid_temperature.degrees_per_tick,
            current=self._settings.lid_temperature.starting,
        )
        self._plate_temperature = TemperatureWithHold(
            per_tick=self._settings.plate_temperature.degrees_per_tick,
            current=self._settings.plate_temperature.starting,
        )
        self.lid_status = ThermocyclerLidStatus.OPEN
        self.plate_volume = util.OptionalValue[float]()
        self.plate_ramp_rate = util.OptionalValue[float]()

    def _handle(self, command: Command) -> Optional[str]:  # noqa: C901
        """
        Handle a command.

        TODO: AL 20210218 create dispatch map and remove 'noqa(C901)'
        """
        logger.info(f"Got command {command}")
        if command.gcode == GCODE.OPEN_LID:
            self.lid_status = ThermocyclerLidStatus.OPEN
        elif command.gcode == GCODE.CLOSE_LID:
            self.lid_status = ThermocyclerLidStatus.CLOSED
        elif command.gcode == GCODE.GET_LID_STATUS:
            return f"Lid:{self.lid_status}"
        elif command.gcode == GCODE.SET_LID_TEMP:
            temperature = command.params["S"]
            assert isinstance(
                temperature, float
            ), f"invalid temperature '{temperature}'"
            self._lid_temperature.set_target(temperature)
        elif command.gcode == GCODE.GET_LID_TEMP:
            res = (
                f"T:{util.OptionalValue(self._lid_temperature.target)} "
                f"C:{self._lid_temperature.current} "
                f"H:none Total_H:none"
            )
            self._lid_temperature.tick()
            return res
        elif command.gcode == GCODE.EDIT_PID_PARAMS:
            pass
        elif command.gcode == GCODE.SET_PLATE_TEMP:
            for prefix, value in command.params.items():
                assert isinstance(value, float), f"invalid value '{value}'"
                if prefix == "S":
                    self._plate_temperature.set_target(value)
                elif prefix == "V":
                    self.plate_volume.val = value
                elif prefix == "H":
                    self._plate_temperature.set_hold(value)
        elif command.gcode == GCODE.GET_PLATE_TEMP:
            plate_target = util.OptionalValue(self._plate_temperature.target)
            plate_current = self._plate_temperature.current
            plate_time_remaining = util.OptionalValue(
                self._plate_temperature.time_remaining
            )
            plate_total_hold_time = util.OptionalValue(
                self._plate_temperature.total_hold
            )

            res = (
                f"T:{plate_target} "
                f"C:{plate_current} "
                f"H:{plate_time_remaining} "
                f"Total_H:{plate_total_hold_time} "
            )
            self._plate_temperature.tick()
            return res
        elif command.gcode == GCODE.SET_RAMP_RATE:
            self.plate_ramp_rate.val = command.params["S"]
        elif command.gcode == GCODE.DEACTIVATE_ALL:
            self._plate_temperature.deactivate(temperature=util.TEMPERATURE_ROOM)
            self._lid_temperature.deactivate(temperature=util.TEMPERATURE_ROOM)
        elif command.gcode == GCODE.DEACTIVATE_LID:
            self._lid_temperature.deactivate(temperature=util.TEMPERATURE_ROOM)
        elif command.gcode == GCODE.DEACTIVATE_BLOCK:
            self._plate_temperature.deactivate(temperature=util.TEMPERATURE_ROOM)
        elif command.gcode == GCODE.DEVICE_INFO:
            return (
                f"serial:{self._settings.serial_number} "
                f"model:{self._settings.model} "
                f"version:{self._settings.version}"
            )
        return None

    @staticmethod
    def get_terminator() -> bytes:
        return b"\r\n"



"""backends.types - wrapper types for api/backend interaction"""

from enum import Enum, auto


class HWStopCondition(Enum):
    none = auto()
    limit_switch = auto()
    sync_line = auto()
    encoder_position = auto()
    gripper_force = auto()
    stall = auto()
    ignore_stalls = auto()
    limit_switch_backoff = auto()



"""Command models to wait for heating a Thermocycler's lid."""
from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler


WaitForLidTemperatureCommandType = Literal["thermocycler/waitForLidTemperature"]


class WaitForLidTemperatureParams(BaseModel):
    """Input parameters to wait for Thermocycler's lid temperature."""

    moduleId: str = Field(..., description="Unique ID of the Thermocycler Module.")


class WaitForLidTemperatureResult(BaseModel):
    """Result data from wait for Thermocycler's lid temperature."""


class WaitForLidTemperatureImpl(
    AbstractCommandImpl[
        WaitForLidTemperatureParams, SuccessData[WaitForLidTemperatureResult]
    ]
):
    """Execution implementation of Thermocycler's wait for lid temperature command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self,
        params: WaitForLidTemperatureParams,
    ) -> SuccessData[WaitForLidTemperatureResult]:
        """Wait for a Thermocycler's lid temperature."""
        thermocycler_state = self._state_view.modules.get_thermocycler_module_substate(
            params.moduleId
        )

        # Raises error if no target temperature
        thermocycler_state.get_target_lid_temperature()

        thermocycler_hardware = self._equipment.get_module_hardware_api(
            thermocycler_state.module_id
        )

        if thermocycler_hardware is not None:
            await thermocycler_hardware.wait_for_lid_target()

        return SuccessData(
            public=WaitForLidTemperatureResult(),
        )


class WaitForLidTemperature(
    BaseCommand[
        WaitForLidTemperatureParams, WaitForLidTemperatureResult, ErrorOccurrence
    ]
):
    """A command to wait for a Thermocycler's lid temperature."""

    commandType: WaitForLidTemperatureCommandType = "thermocycler/waitForLidTemperature"
    params: WaitForLidTemperatureParams
    result: Optional[WaitForLidTemperatureResult] = None

    _ImplementationCls: Type[WaitForLidTemperatureImpl] = WaitForLidTemperatureImpl


class WaitForLidTemperatureCreate(BaseCommandCreate[WaitForLidTemperatureParams]):
    """A request to create Thermocycler's wait for lid temperature command."""

    commandType: WaitForLidTemperatureCommandType = "thermocycler/waitForLidTemperature"
    params: WaitForLidTemperatureParams

    _CommandCls: Type[WaitForLidTemperature] = WaitForLidTemperature



"""Command models to close the lid on an Absorbance Reader."""
from __future__ import annotations
from typing import Optional, Literal, TYPE_CHECKING
from typing_extensions import Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence
from ...errors import CannotPerformModuleAction
from opentrons.protocol_engine.types import AddressableAreaLocation

from ...state.update_types import StateUpdate


from opentrons.drivers.types import AbsorbanceReaderLidStatus

if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView
    from opentrons.protocol_engine.execution import (
        EquipmentHandler,
        LabwareMovementHandler,
    )


CloseLidCommandType = Literal["absorbanceReader/closeLid"]


class CloseLidParams(BaseModel):
    """Input parameters to close the lid on an absorbance reading."""

    moduleId: str = Field(..., description="Unique ID of the absorbance reader.")


class CloseLidResult(BaseModel):
    """Result data from closing the lid on an aborbance reading."""


class CloseLidImpl(AbstractCommandImpl[CloseLidParams, SuccessData[CloseLidResult]]):
    """Execution implementation of closing the lid on an Absorbance Reader."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        labware_movement: LabwareMovementHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment
        self._labware_movement = labware_movement

    async def execute(self, params: CloseLidParams) -> SuccessData[CloseLidResult]:
        """Execute the close lid command."""
        state_update = StateUpdate()
        mod_substate = self._state_view.modules.get_absorbance_reader_substate(
            module_id=params.moduleId
        )

        hardware_lid_status = AbsorbanceReaderLidStatus.OFF
        if not self._state_view.config.use_virtual_modules:
            abs_reader = self._equipment.get_module_hardware_api(mod_substate.module_id)
            if abs_reader is not None:
                hardware_lid_status = await abs_reader.get_current_lid_status()
            else:
                raise CannotPerformModuleAction(
                    "Could not reach the Hardware API for Opentrons Plate Reader Module."
                )

        if hardware_lid_status is AbsorbanceReaderLidStatus.ON:
            # The lid is already physically ON, so we can no-op physically closing it
            state_update.set_absorbance_reader_lid(
                module_id=mod_substate.module_id, is_lid_on=True
            )
        else:
            # Allow propagation of ModuleNotAttachedError.
            _ = self._equipment.get_module_hardware_api(mod_substate.module_id)

            lid_definition = (
                self._state_view.labware.get_absorbance_reader_lid_definition()
            )

            current_location = self._state_view.modules.absorbance_reader_dock_location(
                params.moduleId
            )

            # we need to move the lid onto the module reader
            absorbance_model = self._state_view.modules.get_requested_model(
                params.moduleId
            )
            assert absorbance_model is not None
            new_location = AddressableAreaLocation(
                addressableAreaName=self._state_view.modules.ensure_and_convert_module_fixture_location(
                    deck_slot=self._state_view.modules.get_location(
                        params.moduleId
                    ).slotName,
                    model=absorbance_model,
                )
            )

            # The lid's labware definition stores gripper offsets for itself in the
            # space normally meant for offsets for labware stacked atop it.
            lid_gripper_offsets = self._state_view.labware.get_child_gripper_offsets(
                labware_definition=lid_definition,
                slot_name=None,
            )
            if lid_gripper_offsets is None:
                raise ValueError(
                    "Gripper Offset values for Absorbance Reader Lid labware must not be None."
                )

            await self._labware_movement.move_labware_with_gripper(
                labware_definition=lid_definition,
                current_location=current_location,
                new_location=new_location,
                user_offset_data=lid_gripper_offsets,
                post_drop_slide_offset=None,
            )
            state_update.set_absorbance_reader_lid(
                module_id=mod_substate.module_id,
                is_lid_on=True,
            )

        return SuccessData(
            public=CloseLidResult(),
            state_update=state_update,
        )


class CloseLid(BaseCommand[CloseLidParams, CloseLidResult, ErrorOccurrence]):
    """A command to close the lid on an Absorbance Reader."""

    commandType: CloseLidCommandType = "absorbanceReader/closeLid"
    params: CloseLidParams
    result: Optional[CloseLidResult] = None

    _ImplementationCls: Type[CloseLidImpl] = CloseLidImpl


class CloseLidCreate(BaseCommandCreate[CloseLidParams]):
    """A request to execute an Absorbance Reader close lid command."""

    commandType: CloseLidCommandType = "absorbanceReader/closeLid"
    params: CloseLidParams

    _CommandCls: Type[CloseLid] = CloseLid



"""Touch tip command request, result, and implementation models."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Any

from typing_extensions import Literal
from pydantic import Field
from pydantic.json_schema import SkipJsonSchema

from opentrons.types import Point

from ..errors import (
    TouchTipDisabledError,
    TouchTipIncompatibleArgumentsError,
    LabwareIsTipRackError,
)
from ..types import DeckPoint
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
    DefinedErrorData,
)
from .pipetting_common import (
    PipetteIdMixin,
)
from .movement_common import (
    WellLocationMixin,
    DestinationPositionResult,
    StallOrCollisionError,
    move_to_well,
)

if TYPE_CHECKING:
    from ..execution import MovementHandler, GantryMover
    from ..state.state import StateView
    from ..resources.model_utils import ModelUtils


TouchTipCommandType = Literal["touchTip"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class TouchTipParams(PipetteIdMixin, WellLocationMixin):
    """Payload needed to touch a pipette tip the sides of a specific well."""

    radius: float = Field(
        1.0,
        description=(
            "The proportion of the target well's radius the pipette tip will move towards."
        ),
    )

    mmFromEdge: float | SkipJsonSchema[None] = Field(
        None,
        description="Offset away from the the well edge, in millimeters."
        "Incompatible when a radius is included as a non 1.0 value.",
        json_schema_extra=_remove_default,
    )

    speed: float | SkipJsonSchema[None] = Field(
        None,
        description=(
            "Override the travel speed in mm/s."
            " This controls the straight linear speed of motion."
        ),
        json_schema_extra=_remove_default,
    )


class TouchTipResult(DestinationPositionResult):
    """Result data from the execution of a TouchTip."""

    pass


class TouchTipImplementation(
    AbstractCommandImpl[
        TouchTipParams,
        SuccessData[TouchTipResult] | DefinedErrorData[StallOrCollisionError],
    ]
):
    """Touch tip command implementation."""

    def __init__(
        self,
        state_view: StateView,
        movement: MovementHandler,
        gantry_mover: GantryMover,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._movement = movement
        self._gantry_mover = gantry_mover
        self._model_utils = model_utils

    async def execute(
        self, params: TouchTipParams
    ) -> SuccessData[TouchTipResult] | DefinedErrorData[StallOrCollisionError]:
        """Touch tip to sides of a well using the requested pipette."""
        pipette_id = params.pipetteId
        labware_id = params.labwareId
        well_name = params.wellName

        if params.radius != 1.0 and params.mmFromEdge is not None:
            raise TouchTipIncompatibleArgumentsError(
                "Cannot use mmFromEdge with a radius that is not 1.0"
            )

        if self._state_view.labware.get_has_quirk(labware_id, "touchTipDisabled"):
            raise TouchTipDisabledError(
                f"Touch tip not allowed on labware {labware_id}"
            )

        if self._state_view.labware.is_tiprack(labware_id):
            raise LabwareIsTipRackError("Cannot touch tip on tip rack")

        center_result = await move_to_well(
            movement=self._movement,
            model_utils=self._model_utils,
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=params.wellLocation,
        )
        if isinstance(center_result, DefinedErrorData):
            return center_result

        touch_speed = self._state_view.pipettes.get_movement_speed(
            pipette_id, params.speed
        )

        mm_from_edge = params.mmFromEdge if params.mmFromEdge is not None else 0
        touch_waypoints = self._state_view.motion.get_touch_tip_waypoints(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            radius=params.radius,
            mm_from_edge=mm_from_edge,
            center_point=Point(
                center_result.public.position.x,
                center_result.public.position.y,
                center_result.public.position.z,
            ),
        )

        final_point = await self._gantry_mover.move_to(
            pipette_id=pipette_id,
            waypoints=touch_waypoints,
            speed=touch_speed,
        )
        final_deck_point = DeckPoint.model_construct(
            x=final_point.x, y=final_point.y, z=final_point.z
        )
        state_update = center_result.state_update.set_pipette_location(
            pipette_id=pipette_id,
            new_labware_id=labware_id,
            new_well_name=well_name,
            new_deck_point=final_deck_point,
        )

        return SuccessData(
            public=TouchTipResult(position=final_deck_point),
            state_update=state_update,
        )


class TouchTip(BaseCommand[TouchTipParams, TouchTipResult, StallOrCollisionError]):
    """Touch up tip command model."""

    commandType: TouchTipCommandType = "touchTip"
    params: TouchTipParams
    result: Optional[TouchTipResult] = None

    _ImplementationCls: Type[TouchTipImplementation] = TouchTipImplementation


class TouchTipCreate(BaseCommandCreate[TouchTipParams]):
    """Touch tip command creation request model."""

    commandType: TouchTipCommandType = "touchTip"
    params: TouchTipParams

    _CommandCls: Type[TouchTip] = TouchTip



"""Protocol Engine types to do with liquid level detection."""

from __future__ import annotations
from dataclasses import dataclass
from datetime import datetime
from typing import Optional, List, Any
from pydantic import BaseModel, model_serializer, model_validator


class SimulatedProbeResult(BaseModel):
    """A sentinel value to substitute for the resulting volume/height of a liquid probe during simulation."""

    operations_after_probe: List[float] = []
    net_liquid_exchanged_after_probe: float = 0.0

    @model_serializer
    def serialize_model(self) -> str:
        """Serialize instances of this class as a string."""
        return "SimulatedProbeResult"

    @model_validator(mode="before")
    @classmethod
    def validate_model(cls, data: object) -> Any:
        """Handle deserializing from a simulated probe result."""
        if isinstance(data, str) and data == "SimulatedProbeResult":
            return {}
        return data

    def __add__(
        self, other: float | int | SimulatedProbeResult
    ) -> float | SimulatedProbeResult:
        """Bypass addition and just return self."""
        return self

    def __sub__(
        self, other: float | int | SimulatedProbeResult
    ) -> float | SimulatedProbeResult:
        """Bypass subtraction and just return self."""
        return self

    def __radd__(
        self, other: float | int | SimulatedProbeResult
    ) -> float | SimulatedProbeResult:
        """Bypass addition and just return self."""
        return self

    def __rsub__(
        self, other: float | int | SimulatedProbeResult
    ) -> float | SimulatedProbeResult:
        """Bypass subtraction and just return self."""
        return self

    def __mul__(
        self, other: float | int | SimulatedProbeResult
    ) -> float | SimulatedProbeResult:
        """Bypass multiplication and just return self."""
        return self

    def __rmul__(
        self, other: float | int | SimulatedProbeResult
    ) -> float | SimulatedProbeResult:
        """Bypass multiplication and just return self."""
        return self

    def __truediv__(
        self, other: float | int | SimulatedProbeResult
    ) -> float | SimulatedProbeResult:
        """Bypass division and just return self."""
        return self

    def __rtruediv__(
        self, other: float | int | SimulatedProbeResult
    ) -> float | SimulatedProbeResult:
        """Bypass division and just return self."""
        return self

    def __pow__(
        self, other: float | int | SimulatedProbeResult
    ) -> float | SimulatedProbeResult:
        """Bypass exponent math and just return self."""
        return self

    def __rpow__(
        self, other: float | int | SimulatedProbeResult
    ) -> float | SimulatedProbeResult:
        """Bypass exponent math and just return self."""
        return self

    def __mod__(
        self, other: float | int | SimulatedProbeResult
    ) -> float | SimulatedProbeResult:
        """Bypass modulus and just return self."""
        return self

    def __rmod__(
        self, other: float | int | SimulatedProbeResult
    ) -> float | SimulatedProbeResult:
        """Bypass modulus and just return self."""
        return self

    def __floordiv__(
        self, other: float | int | SimulatedProbeResult
    ) -> float | SimulatedProbeResult:
        """Bypass floor division and just return self."""
        return self

    def __rfloordiv__(
        self, other: float | int | SimulatedProbeResult
    ) -> float | SimulatedProbeResult:
        """Bypass floor division and just return self."""
        return self

    def __gt__(self, other: float | int | SimulatedProbeResult) -> bool:
        """Bypass 'greater than' and just return self."""
        return True

    def __lt__(self, other: float | int | SimulatedProbeResult) -> bool:
        """Bypass 'less than' and just return self."""
        return False

    def __ge__(self, other: float | int | SimulatedProbeResult) -> bool:
        """Bypass 'greater than or eaqual to' and just return self."""
        return True

    def __le__(self, other: float | int | SimulatedProbeResult) -> bool:
        """Bypass 'less than or equal to' and just return self."""
        return False

    def __eq__(self, other: object) -> bool:
        """A SimulatedProbeResult should only be equal to the same instance of its class."""
        if not isinstance(other, SimulatedProbeResult):
            return False
        return self is other

    def __neq__(self, other: object) -> bool:
        """A SimulatedProbeResult should only be equal to the same instance of its class."""
        if not isinstance(other, SimulatedProbeResult):
            return True
        return self is not other

    def simulate_probed_aspirate_dispense(self, volume: float) -> None:
        """Record the current state of aspirate/dispense calls."""
        self.net_liquid_exchanged_after_probe += volume
        self.operations_after_probe.append(volume)


# Work around https://github.com/pydantic/pydantic/issues/6830 - do not change the order of
# this union
LiquidTrackingType = float | SimulatedProbeResult


class LoadedVolumeInfo(BaseModel):
    """A well's liquid volume, initialized by a LoadLiquid, updated by Aspirate and Dispense."""

    volume: LiquidTrackingType | None = None
    last_loaded: datetime
    operations_since_load: int


class ProbedHeightInfo(BaseModel):
    """A well's liquid height, initialized by a LiquidProbe, cleared by Aspirate and Dispense."""

    height: LiquidTrackingType | None = None
    last_probed: datetime


class ProbedVolumeInfo(BaseModel):
    """A well's liquid volume, initialized by a LiquidProbe, updated by Aspirate and Dispense."""

    volume: LiquidTrackingType | None = None
    last_probed: datetime
    operations_since_probe: int


class WellInfoSummary(BaseModel):
    """Payload for a well's liquid info in StateSummary."""

    labware_id: str
    well_name: str
    loaded_volume: Optional[float] = None
    probed_height: LiquidTrackingType | None = None
    probed_volume: LiquidTrackingType | None = None


@dataclass
class WellLiquidInfo:
    """Tracked and sensed information about liquid in a well."""

    probed_height: Optional[ProbedHeightInfo]
    loaded_volume: Optional[LoadedVolumeInfo]
    probed_volume: Optional[ProbedVolumeInfo]



from typing import Dict, Optional
from typing_extensions import Protocol
from ..types import DoorState, StatusBarState, EstopState
from .event_sourcer import EventSourcer


class ChassisAccessoryManager(EventSourcer, Protocol):
    """Protocol specifying control of non-motion peripherals on the robot."""

    @property
    def door_state(self) -> DoorState:
        """The current state of the machine's door."""
        ...

    async def set_lights(
        self,
        button: Optional[bool] = None,
        rails: Optional[bool] = None,
    ) -> None:
        """Control the robot lights.

        button If specified, turn the button light on (`True`) or
               off (`False`). If not specified, do not change the
               button light.
        rails: If specified, turn the rail lights on (`True`) or
               off (`False`). If not specified, do not change the
               rail lights.
        """
        ...

    async def get_lights(self) -> Dict[str, bool]:
        """Return the current status of the robot lights.

        :returns: A dict of the lights: `{'button': bool, 'rails': bool}`
        """
        ...

    async def identify(self, duration_s: int = 5) -> None:
        """Run a routine to identify the robot.

        duration_s: The duration to blink for, in seconds.
        """
        ...

    async def set_status_bar_state(self, state: StatusBarState) -> None:
        """Control the status bar to indicate robot state.

        state: The state to set the robot to. Some states are transient
        and will implicitly revert back to the previous state after a short
        action, while others"""
        ...

    async def set_status_bar_enabled(self, enabled: bool) -> None:
        """Enable or disable the status bar entirely.

        enabled: True to turn the status bar on, false to turn it off."""

    def get_status_bar_state(self) -> StatusBarState:
        """Get the current status bar state.

        :returns: The current status bar state enumeration."""
        ...

    def get_estop_state(self) -> EstopState:
        """Get the current Estop state.

        If the Estop is not supported on this robot, this will always return Disengaged.

        :returns: The current Estop state."""
        ...



from __future__ import annotations

import asyncio
import logging
from typing import Dict, Optional

from opentrons.hardware_control.modules.types import (
    ModuleDisconnectedCallback,
    TemperatureStatus,
)
from opentrons.hardware_control.poller import Reader, Poller
from typing_extensions import Final
from opentrons.drivers.types import Temperature
from opentrons.drivers.temp_deck import (
    SimulatingDriver,
    AbstractTempDeckDriver,
    TempDeckDriver,
)
from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.hardware_control.execution_manager import ExecutionManager
from opentrons.hardware_control.modules import update, mod_abc, types, errors

log = logging.getLogger(__name__)

TEMP_POLL_INTERVAL_SECS = 1.0
SIM_TEMP_POLL_INTERVAL_SECS = TEMP_POLL_INTERVAL_SECS / 20.0


class TempDeck(mod_abc.AbstractModule):
    """Hardware control interface for an attached Temperature Module."""

    MODULE_TYPE = types.ModuleType["TEMPERATURE"]
    FIRST_GEN2_REVISION = 20

    @classmethod
    async def build(
        cls,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        hw_control_loop: asyncio.AbstractEventLoop,
        poll_interval_seconds: Optional[float] = None,
        simulating: bool = False,
        sim_model: Optional[str] = None,
        sim_serial_number: Optional[str] = None,
        disconnected_callback: ModuleDisconnectedCallback = None,
    ) -> "TempDeck":
        """
        Build a TempDeck

        Args:
            port: The port to connect to
            usb_port: USB Port
            execution_manager: Execution manager.
            hw_control_loop: The event loop running in the hardware control thread.
            poll_interval_seconds: Poll interval override.
            simulating: whether to build a simulating driver
            sim_model: The model name used by simulator
            disconnected_callback: Callback to inform the module controller that the device was disconnected

        Returns:
            Tempdeck instance
        """
        driver: AbstractTempDeckDriver
        if not simulating:
            driver = await TempDeckDriver.create(port=port, loop=hw_control_loop)
            poll_interval_seconds = poll_interval_seconds or TEMP_POLL_INTERVAL_SECS
        else:
            driver = SimulatingDriver(
                sim_model=sim_model, serial_number=sim_serial_number
            )
            poll_interval_seconds = poll_interval_seconds or SIM_TEMP_POLL_INTERVAL_SECS

        reader = TempDeckReader(driver=driver)
        poller = Poller(reader=reader, interval=poll_interval_seconds)
        module = cls(
            port=port,
            usb_port=usb_port,
            execution_manager=execution_manager,
            driver=driver,
            reader=reader,
            poller=poller,
            device_info=await driver.get_device_info(),
            hw_control_loop=hw_control_loop,
            disconnected_callback=disconnected_callback,
        )

        try:
            await poller.start()
        except Exception:
            log.exception(f"First read of Temperature Module on port {port} failed")

        return module

    def __init__(
        self,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        driver: AbstractTempDeckDriver,
        reader: TempDeckReader,
        poller: Poller,
        device_info: Dict[str, str],
        hw_control_loop: asyncio.AbstractEventLoop,
        disconnected_callback: ModuleDisconnectedCallback = None,
    ) -> None:
        """Constructor"""
        super().__init__(
            port=port,
            usb_port=usb_port,
            hw_control_loop=hw_control_loop,
            execution_manager=execution_manager,
            disconnected_callback=disconnected_callback,
        )
        self._device_info = device_info
        self._driver = driver
        self._reader = reader
        self._poller = poller

    async def cleanup(self) -> None:
        """Stop the poller task."""
        await self._poller.stop()
        await self._driver.disconnect()

    @classmethod
    def name(cls) -> str:
        return "tempdeck"

    def firmware_prefix(self) -> str:
        """The prefix used for looking up firmware"""
        return "temperature-module"

    def model(self) -> str:
        return self._model_from_revision(self._device_info.get("model"))

    def bootloader(self) -> types.UploadFunction:
        return update.upload_via_avrdude

    async def start_set_temperature(self, celsius: float) -> None:
        """Set the target temperature in degrees Celsius.

        Range: 4 to 95 degrees Celsius (QA tested).

        The internal temp range is -9 to 99 °C, which is limited by the two-digit
        temperature display. Any input outside of this range will be clipped
        to the nearest limit.
        """
        await self.wait_for_is_running()
        await self._driver.set_temperature(celsius)
        await self._reader.read()

    async def await_temperature(self, awaiting_temperature: Optional[float]) -> None:
        """Await a target temperature in degrees Celsius.

        Polls Temperature Module's temperature until
        the specified temperature is reached.

        Args:
            temperature: The temperature to wait for.
                If `None` (recommended), the module's target will be used.
                Specifying any value other than the current target
                may produce unpredictable behavior.
        """
        if self.is_simulated:
            return

        await self.wait_for_is_running()
        await self._reader.read()

        async def _await_temperature() -> None:
            if awaiting_temperature is None:
                while self.status != TemperatureStatus.HOLDING:
                    await self._poller.wait_next_poll()
            elif self.status == TemperatureStatus.HEATING:
                while self.temperature < awaiting_temperature:
                    await self._poller.wait_next_poll()
            elif self.status == TemperatureStatus.COOLING:
                while self.temperature > awaiting_temperature:
                    await self._poller.wait_next_poll()

        t = self._loop.create_task(_await_temperature())
        self.make_cancellable(t)
        await t

    async def deactivate(self, must_be_running: bool = True) -> None:
        """Stop heating/cooling and turn off the fan"""
        if must_be_running:
            await self.wait_for_is_running()
        await self._driver.deactivate()
        await self._reader.read()

    @property
    def device_info(self) -> Dict[str, str]:
        return self._device_info

    @property
    def live_data(self) -> types.LiveData:
        data: types.TemperatureModuleData = {
            "currentTemp": self.temperature,
            "targetTemp": self.target,
        }
        return {
            "status": self.status,
            "data": data,
        }

    @property
    def temperature(self) -> float:
        return self._reader.temperature.current

    @property
    def target(self) -> Optional[float]:
        return self._reader.temperature.target

    @property
    def status(self) -> TemperatureStatus:
        return self._get_status(self._reader.temperature)

    @property
    def is_simulated(self) -> bool:
        return isinstance(self._driver, SimulatingDriver)

    async def prep_for_update(self) -> str:
        model = self._device_info and self._device_info.get("model")
        if model in ("temp_deck_v1", "temp_deck_v1.1", "temp_deck_v2"):
            raise errors.UpdateError(
                "This Temperature Module can't be updated."
                "Please contact Opentrons Support."
            )
        await self._poller.stop()
        await self._driver.enter_programming_mode()
        new_port = await update.find_bootloader_port()
        return new_port or self.port

    def has_available_update(self) -> bool:
        """Override of abc implementation to suppress update notifications
        for v1, v1.1, and v2 temperature modules which cannot be updated"""
        if not self._device_info:
            model = None
        else:
            model = self._device_info.get("model")
        if model in {"temp_deck_v1", "temp_deck_v1.1", "temp_deck_v2", None}:
            return False
        return super().has_available_update()

    @staticmethod
    def _get_status(temperature: Temperature) -> TemperatureStatus:
        """
        Determine the status from the temperature.

        Args:
            temperature: A Temperature instance

        Returns:
            The status
        """
        DELTA: Final = 0.7
        status = TemperatureStatus.IDLE
        if temperature.target is not None:
            diff = temperature.target - temperature.current
            if abs(diff) < DELTA:  # To avoid status fluctuation near target
                status = TemperatureStatus.HOLDING
            elif diff < 0:
                status = TemperatureStatus.COOLING
            else:
                status = TemperatureStatus.HEATING
        return status

    @staticmethod
    def _model_from_revision(revision: Optional[str]) -> str:
        """Defines the revision -> model mapping"""
        if not revision or "v" not in revision:
            log.error(f"bad revision: {revision}")
            return "temperatureModuleV1"
        try:
            revision_num = float(revision.split("v")[-1])
        except (ValueError, TypeError):
            # none or corrupt
            log.exception("no revision")
            return "temperatureModuleV1"

        if revision_num < TempDeck.FIRST_GEN2_REVISION:
            return "temperatureModuleV1"
        else:
            return "temperatureModuleV2"


class TempDeckReader(Reader):
    """Reads data from an attached Temperature Module."""

    temperature: Temperature

    def __init__(self, driver: AbstractTempDeckDriver) -> None:
        self.temperature = Temperature(current=25, target=None)
        self._driver = driver

    async def read(self) -> None:
        """Read the module's current and target temperatures."""
        self.temperature = await self._driver.get_temperature()



from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Tuple
from opentrons.drivers.types import (
    ABSMeasurementMode,
    AbsorbanceReaderLidStatus,
    AbsorbanceReaderDeviceState,
    AbsorbanceReaderPlatePresence,
)


class AbstractAbsorbanceReaderDriver(ABC):
    @abstractmethod
    async def connect(self) -> None:
        """Connect to absorbance reader"""
        ...

    @abstractmethod
    async def disconnect(self) -> None:
        """Disconnect from absorbance reader"""
        ...

    @abstractmethod
    async def is_connected(self) -> bool:
        """Check connection to absorbance reader"""
        ...

    @abstractmethod
    async def get_lid_status(self) -> AbsorbanceReaderLidStatus:
        ...

    @abstractmethod
    async def get_available_wavelengths(self) -> List[int]:
        ...

    @abstractmethod
    async def initialize_measurement(
        self,
        wavelengths: List[int],
        mode: ABSMeasurementMode = ABSMeasurementMode.SINGLE,
        reference_wavelength: Optional[int] = None,
    ) -> None:
        """Initialize measurement for the device in single or multi mode for the given wavelengths"""
        ...

    @abstractmethod
    async def get_measurement(self) -> List[List[float]]:
        """Gets one or more measurements based on the current configuration."""
        ...

    @abstractmethod
    async def get_status(self) -> AbsorbanceReaderDeviceState:
        ...

    @abstractmethod
    async def get_device_info(self) -> Dict[str, str]:
        """Get device info"""
        ...

    @abstractmethod
    async def get_uptime(self) -> int:
        """Get device uptime"""
        ...

    @abstractmethod
    async def get_plate_presence(self) -> AbsorbanceReaderPlatePresence:
        """Check if there is a plate in the reader."""
        ...

    @abstractmethod
    async def update_firmware(self, firmware_file_path: str) -> Tuple[bool, str]:
        """Updates the firmware on the device."""
        ...



"""Common movement base models."""

from __future__ import annotations

from typing import Optional, Union, TYPE_CHECKING, Literal, Any

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema

from opentrons_shared_data.errors import ErrorCodes
from opentrons_shared_data.errors.exceptions import StallOrCollisionDetectedError
from ..errors import ErrorOccurrence
from ..types import (
    WellLocation,
    LiquidHandlingWellLocation,
    DeckPoint,
    CurrentWell,
    MovementAxis,
    AddressableOffsetVector,
)
from ..state.update_types import StateUpdate, PipetteLocationUpdate
from .command import SuccessData, DefinedErrorData


if TYPE_CHECKING:
    from ..execution.movement import MovementHandler
    from ..resources.model_utils import ModelUtils


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class WellLocationMixin(BaseModel):
    """Mixin for command requests that take a location that's somewhere in a well."""

    labwareId: str = Field(
        ...,
        description="Identifier of labware to use.",
    )
    wellName: str = Field(
        ...,
        description="Name of well to use in labware.",
    )
    wellLocation: WellLocation = Field(
        default_factory=WellLocation,
        description="Relative well location at which to perform the operation",
    )


class LiquidHandlingWellLocationMixin(BaseModel):
    """Mixin for command requests that take a location that's somewhere in a well."""

    labwareId: str = Field(
        ...,
        description="Identifier of labware to use.",
    )
    wellName: str = Field(
        ...,
        description="Name of well to use in labware.",
    )
    wellLocation: LiquidHandlingWellLocation = Field(
        default_factory=LiquidHandlingWellLocation,
        description="Relative well location at which to perform the operation",
    )


class MovementMixin(BaseModel):
    """Mixin for command requests that move a pipette."""

    minimumZHeight: float | SkipJsonSchema[None] = Field(
        None,
        description=(
            "Optional minimal Z margin in mm."
            " If this is larger than the API's default safe Z margin,"
            " it will make the arc higher. If it's smaller, it will have no effect."
        ),
        json_schema_extra=_remove_default,
    )

    forceDirect: bool = Field(
        False,
        description=(
            "If true, moving from one labware/well to another"
            " will not arc to the default safe z,"
            " but instead will move directly to the specified location."
            " This will also force the `minimumZHeight` param to be ignored."
            " A 'direct' movement is in X/Y/Z simultaneously."
        ),
    )

    speed: float | SkipJsonSchema[None] = Field(
        None,
        description=(
            "Override the travel speed in mm/s."
            " This controls the straight linear speed of motion."
        ),
        json_schema_extra=_remove_default,
    )


class StallOrCollisionError(ErrorOccurrence):
    """Returned when the machine detects that axis encoders are reading a different position than expected.

    All axes are stopped at the point where the error was encountered.

    The next thing to move the machine must account for the robot not having a valid estimate
    of its position. It should be a `home` or `unsafe/updatePositionEstimators`.
    """

    isDefined: bool = True
    errorType: Literal["stallOrCollision"] = "stallOrCollision"

    errorCode: str = ErrorCodes.STALL_OR_COLLISION_DETECTED.value.code
    detail: str = ErrorCodes.STALL_OR_COLLISION_DETECTED.value.detail


class DestinationPositionResult(BaseModel):
    """Mixin for command results that move a pipette."""

    # todo(mm, 2024-08-02): Consider deprecating or redefining this.
    #
    # This is here because opentrons.protocol_engine needed it for internal bookkeeping
    # and, at the time, we didn't have a way to do that without adding this to the
    # public command results. Its usefulness to callers outside
    # opentrons.protocol_engine is questionable because they would need to know which
    # critical point is in play, and I think that can change depending on obscure
    # things like labware quirks.
    position: DeckPoint = Field(
        DeckPoint(x=0, y=0, z=0),
        description=(
            "The (x,y,z) coordinates of the pipette's critical point in deck space"
            " after the move was completed."
        ),
    )


MoveToWellOperationReturn = (
    SuccessData[DestinationPositionResult] | DefinedErrorData[StallOrCollisionError]
)


async def move_to_well(
    movement: MovementHandler,
    model_utils: ModelUtils,
    pipette_id: str,
    labware_id: str,
    well_name: str,
    well_location: Optional[Union[WellLocation, LiquidHandlingWellLocation]] = None,
    current_well: Optional[CurrentWell] = None,
    force_direct: bool = False,
    minimum_z_height: Optional[float] = None,
    speed: Optional[float] = None,
    operation_volume: Optional[float] = None,
) -> MoveToWellOperationReturn:
    """Execute a move to well microoperation."""
    try:
        position = await movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            current_well=current_well,
            force_direct=force_direct,
            minimum_z_height=minimum_z_height,
            speed=speed,
            operation_volume=operation_volume,
        )
    except StallOrCollisionDetectedError as e:
        return DefinedErrorData(
            public=StallOrCollisionError(
                id=model_utils.generate_id(),
                createdAt=model_utils.get_timestamp(),
                wrappedErrors=[
                    ErrorOccurrence.from_failed(
                        id=model_utils.generate_id(),
                        createdAt=model_utils.get_timestamp(),
                        error=e,
                    )
                ],
            ),
            state_update=StateUpdate().clear_all_pipette_locations(),
        )
    else:
        deck_point = DeckPoint.model_construct(x=position.x, y=position.y, z=position.z)
        return SuccessData(
            public=DestinationPositionResult(
                position=deck_point,
            ),
            state_update=StateUpdate().set_pipette_location(
                pipette_id=pipette_id,
                new_labware_id=labware_id,
                new_well_name=well_name,
                new_deck_point=deck_point,
            ),
        )


async def move_relative(
    movement: MovementHandler,
    model_utils: ModelUtils,
    pipette_id: str,
    axis: MovementAxis,
    distance: float,
) -> SuccessData[DestinationPositionResult] | DefinedErrorData[StallOrCollisionError]:
    """Move by a fixed displacement from the current position."""
    try:
        position = await movement.move_relative(pipette_id, axis, distance)
    except StallOrCollisionDetectedError as e:
        return DefinedErrorData(
            public=StallOrCollisionError(
                id=model_utils.generate_id(),
                createdAt=model_utils.get_timestamp(),
                wrappedErrors=[
                    ErrorOccurrence.from_failed(
                        id=model_utils.generate_id(),
                        createdAt=model_utils.get_timestamp(),
                        error=e,
                    )
                ],
            ),
            state_update=StateUpdate().clear_all_pipette_locations(),
        )
    else:
        deck_point = DeckPoint.model_construct(x=position.x, y=position.y, z=position.z)
        return SuccessData(
            public=DestinationPositionResult(
                position=deck_point,
            ),
            state_update=StateUpdate().set_pipette_location(
                pipette_id=pipette_id, new_deck_point=deck_point
            ),
        )


async def move_to_addressable_area(
    movement: MovementHandler,
    model_utils: ModelUtils,
    pipette_id: str,
    addressable_area_name: str,
    offset: AddressableOffsetVector,
    force_direct: bool = False,
    minimum_z_height: float | None = None,
    speed: float | None = None,
    stay_at_highest_possible_z: bool = False,
    ignore_tip_configuration: bool | None = True,
    highest_possible_z_extra_offset: float | None = None,
) -> SuccessData[DestinationPositionResult] | DefinedErrorData[StallOrCollisionError]:
    """Move to an addressable area identified by name."""
    try:
        x, y, z = await movement.move_to_addressable_area(
            pipette_id=pipette_id,
            addressable_area_name=addressable_area_name,
            offset=offset,
            force_direct=force_direct,
            minimum_z_height=minimum_z_height,
            speed=speed,
            stay_at_highest_possible_z=stay_at_highest_possible_z,
            ignore_tip_configuration=ignore_tip_configuration,
            highest_possible_z_extra_offset=highest_possible_z_extra_offset,
        )
    except StallOrCollisionDetectedError as e:
        return DefinedErrorData(
            public=StallOrCollisionError(
                id=model_utils.generate_id(),
                createdAt=model_utils.get_timestamp(),
                wrappedErrors=[
                    ErrorOccurrence.from_failed(
                        id=model_utils.generate_id(),
                        createdAt=model_utils.get_timestamp(),
                        error=e,
                    )
                ],
            ),
            state_update=StateUpdate()
            .clear_all_pipette_locations()
            .set_addressable_area_used(addressable_area_name=addressable_area_name),
        )
    else:
        deck_point = DeckPoint.model_construct(x=x, y=y, z=z)
        return SuccessData(
            public=DestinationPositionResult(position=deck_point),
            state_update=StateUpdate()
            .set_pipette_location(
                pipette_id=pipette_id,
                new_addressable_area_name=addressable_area_name,
                new_deck_point=deck_point,
            )
            .set_addressable_area_used(addressable_area_name=addressable_area_name),
        )


async def move_to_coordinates(
    movement: MovementHandler,
    model_utils: ModelUtils,
    pipette_id: str,
    deck_coordinates: DeckPoint,
    direct: bool,
    additional_min_travel_z: float | None,
    speed: float | None = None,
) -> SuccessData[DestinationPositionResult] | DefinedErrorData[StallOrCollisionError]:
    """Move to a set of coordinates."""
    try:
        x, y, z = await movement.move_to_coordinates(
            pipette_id=pipette_id,
            deck_coordinates=deck_coordinates,
            direct=direct,
            additional_min_travel_z=additional_min_travel_z,
            speed=speed,
        )
    except StallOrCollisionDetectedError as e:
        return DefinedErrorData(
            public=StallOrCollisionError(
                id=model_utils.generate_id(),
                createdAt=model_utils.get_timestamp(),
                wrappedErrors=[
                    ErrorOccurrence.from_failed(
                        id=model_utils.generate_id(),
                        createdAt=model_utils.get_timestamp(),
                        error=e,
                    )
                ],
            ),
            state_update=StateUpdate().clear_all_pipette_locations(),
        )
    else:
        deck_point = DeckPoint.model_construct(x=x, y=y, z=z)

        return SuccessData(
            public=DestinationPositionResult(position=DeckPoint(x=x, y=y, z=z)),
            state_update=StateUpdate(
                pipette_location=PipetteLocationUpdate(
                    pipette_id=pipette_id,
                    new_location=None,
                    new_deck_point=deck_point,
                )
            ),
        )



"""Command models to execute a Thermocycler profile."""
from __future__ import annotations
from typing import List, Optional, TYPE_CHECKING, overload, Union, Any
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema

from opentrons.hardware_control.modules.types import ThermocyclerStep, ThermocyclerCycle

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler
    from opentrons.protocol_engine.state.module_substates.thermocycler_module_substate import (
        ThermocyclerModuleSubState,
    )


RunExtendedProfileCommandType = Literal["thermocycler/runExtendedProfile"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class ProfileStep(BaseModel):
    """An individual step in a Thermocycler extended profile."""

    celsius: float = Field(..., description="Target temperature in °C.")
    holdSeconds: float = Field(
        ..., description="Time to hold target temperature in seconds."
    )


class ProfileCycle(BaseModel):
    """An individual cycle in a Thermocycler extended profile."""

    steps: List[ProfileStep] = Field(..., description="Steps to repeat.")
    repetitions: int = Field(..., description="Number of times to repeat the steps.")


class RunExtendedProfileParams(BaseModel):
    """Input parameters for an individual Thermocycler profile step."""

    moduleId: str = Field(..., description="Unique ID of the Thermocycler.")
    profileElements: List[Union[ProfileStep, ProfileCycle]] = Field(
        ...,
        description="Elements of the profile. Each can be either a step or a cycle.",
    )
    blockMaxVolumeUl: float | SkipJsonSchema[None] = Field(
        None,
        description="Amount of liquid in uL of the most-full well"
        " in labware loaded onto the thermocycler.",
        json_schema_extra=_remove_default,
    )


class RunExtendedProfileResult(BaseModel):
    """Result data from running a Thermocycler profile."""


def _transform_profile_step(
    step: ProfileStep, thermocycler_state: ThermocyclerModuleSubState
) -> ThermocyclerStep:
    return ThermocyclerStep(
        temperature=thermocycler_state.validate_target_block_temperature(step.celsius),
        hold_time_seconds=step.holdSeconds,
    )


@overload
def _transform_profile_element(
    element: ProfileStep, thermocycler_state: ThermocyclerModuleSubState
) -> ThermocyclerStep:
    ...


@overload
def _transform_profile_element(
    element: ProfileCycle, thermocycler_state: ThermocyclerModuleSubState
) -> ThermocyclerCycle:
    ...


def _transform_profile_element(
    element: Union[ProfileStep, ProfileCycle],
    thermocycler_state: ThermocyclerModuleSubState,
) -> Union[ThermocyclerStep, ThermocyclerCycle]:
    if isinstance(element, ProfileStep):
        return _transform_profile_step(element, thermocycler_state)
    else:
        return ThermocyclerCycle(
            steps=[
                _transform_profile_step(step, thermocycler_state)
                for step in element.steps
            ],
            repetitions=element.repetitions,
        )


class RunExtendedProfileImpl(
    AbstractCommandImpl[RunExtendedProfileParams, SuccessData[RunExtendedProfileResult]]
):
    """Execution implementation of a Thermocycler's run profile command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self, params: RunExtendedProfileParams
    ) -> SuccessData[RunExtendedProfileResult]:
        """Run a Thermocycler profile."""
        thermocycler_state = self._state_view.modules.get_thermocycler_module_substate(
            params.moduleId
        )
        thermocycler_hardware = self._equipment.get_module_hardware_api(
            thermocycler_state.module_id
        )

        profile = [
            _transform_profile_element(element, thermocycler_state)
            for element in params.profileElements
        ]
        target_volume: Optional[float]
        if params.blockMaxVolumeUl is not None:
            target_volume = thermocycler_state.validate_max_block_volume(
                params.blockMaxVolumeUl
            )
        else:
            target_volume = None

        if thermocycler_hardware is not None:
            # TODO(jbl 2022-06-27) hardcoded constant 1 for `repetitions` should be
            #  moved from HardwareControlAPI to the Python ProtocolContext
            await thermocycler_hardware.execute_profile(
                profile=profile, volume=target_volume
            )

        return SuccessData(
            public=RunExtendedProfileResult(),
        )


class RunExtendedProfile(
    BaseCommand[RunExtendedProfileParams, RunExtendedProfileResult, ErrorOccurrence]
):
    """A command to execute a Thermocycler profile run."""

    commandType: RunExtendedProfileCommandType = "thermocycler/runExtendedProfile"
    params: RunExtendedProfileParams
    result: Optional[RunExtendedProfileResult] = None

    _ImplementationCls: Type[RunExtendedProfileImpl] = RunExtendedProfileImpl


class RunExtendedProfileCreate(BaseCommandCreate[RunExtendedProfileParams]):
    """A request to execute a Thermocycler profile run."""

    commandType: RunExtendedProfileCommandType = "thermocycler/runExtendedProfile"
    params: RunExtendedProfileParams

    _CommandCls: Type[RunExtendedProfile] = RunExtendedProfile



from __future__ import annotations

import asyncio
import contextlib
from concurrent.futures.thread import ThreadPoolExecutor
from functools import partial
from typing import Optional, AsyncGenerator, Union
from typing_extensions import Literal

from serial import Serial, serial_for_url  # type: ignore[import-untyped]

TimeoutProperties = Union[Literal["write_timeout"], Literal["timeout"]]


class AsyncSerial:
    """Async wrapper around Serial."""

    @classmethod
    async def create(
        cls,
        port: str,
        baud_rate: int,
        timeout: Optional[float] = None,
        write_timeout: Optional[float] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        reset_buffer_before_write: bool = False,
    ) -> AsyncSerial:
        """
        Create an AsyncSerial instance.

        Args:
            port: url or port name
            baud_rate: the baud rate
            timeout: optional timeout in seconds
            write_timeout: optional write timeout in seconds
            loop: optional event loop. if None get_running_loop will be used
            reset_buffer_before_write: reset the serial input buffer before
             writing to it
        """
        loop = loop or asyncio.get_running_loop()
        executor = ThreadPoolExecutor(max_workers=1)
        serial = await loop.run_in_executor(
            executor=executor,
            func=partial(
                serial_for_url,
                url=port,
                baudrate=baud_rate,
                timeout=timeout,
                write_timeout=write_timeout,
            ),
        )
        return cls(
            serial=serial,
            executor=executor,
            loop=loop,
            reset_buffer_before_write=reset_buffer_before_write,
        )

    def __init__(
        self,
        serial: Serial,
        executor: ThreadPoolExecutor,
        loop: asyncio.AbstractEventLoop,
        reset_buffer_before_write: bool,
    ) -> None:
        """
        Constructor

        Args:
            serial: connected Serial object
            executor: a thread pool executor
            loop: event loop
        """
        self._serial = serial
        self._executor = executor
        self._loop = loop
        self._reset_buffer_before_write = reset_buffer_before_write

    async def read_until(self, match: bytes) -> bytes:
        """
        Read data until match.

        Args:
            match: a sequence of bytes to match
            timeout: optional timeout in seconds. this is a temporary override
                of parameter supplied to `create`

        Returns:
            read data.
        """
        return await self._loop.run_in_executor(
            executor=self._executor,
            func=partial(self._serial.read_until, expected=match),
        )

    async def write(self, data: bytes) -> None:
        """
        Write data

        Args:
            data: data to write.

        Returns:
            None
        """
        await self._loop.run_in_executor(
            executor=self._executor,
            func=lambda: self._sync_write(data=data),
        )

    def _sync_write(self, data: bytes) -> None:
        """
        The synchronous write function

        Args:
            data: data to write.

        Returns:
            None
        """
        if self._reset_buffer_before_write:
            self._serial.reset_input_buffer()
        self._serial.write(data=data)
        self._serial.flush()

    async def open(self) -> None:
        """
        Open the connection.

        Returns: None
        """
        return await self._loop.run_in_executor(
            executor=self._executor, func=self._serial.open
        )

    async def close(self) -> None:
        """
        Close the connection

        Returns: None
        """
        return await self._loop.run_in_executor(
            executor=self._executor, func=self._serial.close
        )

    async def is_open(self) -> bool:
        """
        Check if connection is open.

        Returns: boolean
        """
        return self._serial.is_open is True

    def reset_input_buffer(self) -> None:
        """Reset the input buffer"""
        self._serial.reset_input_buffer()

    def reset_output_buffer(self) -> None:
        """Reset the output buffer"""
        self._serial.reset_output_buffer()

    @contextlib.asynccontextmanager
    async def timeout_override(
        self, timeout_property: TimeoutProperties, timeout: Optional[float]
    ) -> AsyncGenerator[None, None]:
        """Context manager that will temporarily override the default timeout."""
        default_timeout = getattr(self._serial, timeout_property)
        override = timeout is not None and default_timeout != timeout
        try:
            if override:
                await self._loop.run_in_executor(
                    executor=self._executor,
                    func=lambda: setattr(self._serial, timeout_property, timeout),
                )
            yield
        finally:
            if override:
                await self._loop.run_in_executor(
                    executor=self._executor,
                    func=lambda: setattr(
                        self._serial, timeout_property, default_timeout
                    ),
                )



"""File hashing utiliy class."""

from typing import Sequence
import anyio
import unicodedata
from hashlib import md5
from .file_reader_writer import BufferedFile


# TODO (spp: 2024-06-17): move file hasher to utils
class FileHasher:
    """Hashing utility class that hashes a combination of protocol and labware files."""

    @staticmethod
    async def hash(files: Sequence[BufferedFile]) -> str:
        """Sort and hash a sequence of protocol and labware files."""
        return await anyio.to_thread.run_sync(_hash_sync, files)


def _hash_sync(files: Sequence[BufferedFile]) -> str:
    sorted_files = sorted(files, key=lambda x: unicodedata.normalize("NFC", x.name))
    name_content_hasher = md5()
    for file in sorted_files:
        name_hash = md5(file.name.encode("utf-8")).digest()
        contents_hash = md5(file.contents).digest()
        name_content_hasher.update(name_hash + contents_hash)
    return name_content_hasher.hexdigest()



"""Magnetic module sub-state."""

from dataclasses import dataclass
from typing import NewType

from opentrons.protocol_engine.types import MagneticModuleModel, ModuleModel
from opentrons.protocol_engine.errors import EngageHeightOutOfRangeError
from opentrons.hardware_control.modules.magdeck import (
    engage_height_is_in_range,
    OFFSET_TO_LABWARE_BOTTOM as MAGNETIC_MODULE_OFFSET_TO_LABWARE_BOTTOM,
    MAX_ENGAGE_HEIGHT as MAGNETIC_MODULE_MAX_ENGAGE_HEIGHT,
)

MagneticModuleId = NewType("MagneticModuleId", str)


@dataclass(frozen=True)
class MagneticModuleSubState:
    """Magnetic Module specific state.

    Provides calculations and read-only state access
    for an individual loaded Magnetic Module.
    """

    module_id: MagneticModuleId
    model: MagneticModuleModel

    def calculate_magnet_hardware_height(self, mm_from_base: float) -> float:
        """Convert a human-friendly magnet height to be hardware-friendly.

        Args:
            mm_from_base: The height to convert. Measured in how far the tops
                of the magnets are above the labware base plane.

        Returns:
            The same height, with its units and origin point converted
            so that it's suitable to pass to `MagDeck.engage()`.

        Raises:
            EngageHeightOutOfRangeError: If modules of the given model are
                physically incapable of reaching the requested height.
        """
        hardware_units_from_base = (
            mm_from_base * 2
            if self.model == ModuleModel.MAGNETIC_MODULE_V1
            else mm_from_base
        )
        home_to_base_offset = MAGNETIC_MODULE_OFFSET_TO_LABWARE_BOTTOM[self.model]
        hardware_units_from_home = home_to_base_offset + hardware_units_from_base

        if not engage_height_is_in_range(
            model=self.model,
            height=hardware_units_from_home,
        ):
            # TODO(mm, 2022-03-02): This error message probably will not match how
            # the user specified the height. (Hardware units versus mm,
            # home as origin versus labware base as origin.) This may be confusing
            # depending on how it propagates up.
            raise EngageHeightOutOfRangeError(
                f"Invalid engage height for {self.model}:"
                f" {hardware_units_from_home}. Must be"
                f" 0 - {MAGNETIC_MODULE_MAX_ENGAGE_HEIGHT[self.model]}."
            )

        return hardware_units_from_home
